/**
 * *****************************************************************************
 * Copyright (C) 2011 by 52 North Initiative for Geospatial Open Source Software
 * GmbH
 *
 * Contact: Andreas Wytzisk 52 North Initiative for Geospatial Open Source
 * Software GmbH Martin-Luther-King-Weg 24 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 *****************************************************************************
 */
package org.n52.security.common.crypto;

import java.io.UnsupportedEncodingException;
import java.security.MessageDigest;
import java.util.Arrays;
import java.util.Objects;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.apache.commons.codec.binary.Base64;

/**
 * Encapsulation of hashed password values. See
 * http://tools.ietf.org/id/draft-stroeder-hashed-userpassword-values-01.html#rfc.section.2<br/>
 *
 * Algorihtm<br/>
 * ---------<br/>
 * userpasswordvalue = cleartext-password / prefix b64-hashandsalt<br/>
 *
 * prefix = "{" scheme "}"<br/>
 *
 * scheme = %x30-39 / %x41-5A / %x61-7a / %x2D-2F / %x5F ;0-9, A-Z, a-z, "-",
 * ".", "/", or "_"<br/>
 *
 * b64-hashandsalt = &gt;base64 of hashandsalt&lt;<br/>
 *
 * hashandsalt = password-hash salt<br/>
 *
 * password-hash = &gt;digest of cleartext-password salt&lt;<br/>
 *
 * cleartext-password = %x00-FF<br/>
 *
 * salt = %x00-FF<br/>
 *
 *
 * Possible Schemes:<br/>
 * -----------------<br/>
 * {MD5}	MD-5 without salt	[RFC1321]<br/>
 * {SMD5}	salted MD-5 [RFC1321]<br/>
 * {SHA}	SHA-1 without salt	[FIPS-180-4]<br/>
 * {SSHA}	salted SHA-1 [FIPS-180-4]<br/>
 * {SHA256}	SHA-256 without salt	[FIPS-180-4]<br/>
 * {SSHA256}	salted SHA-256 [FIPS-180-4]<br/>
 * {SHA384}	SHA-384 without salt	[FIPS-180-4]<br/>
 * {SSHA384} salted SHA-384 [FIPS-180-4]<br/>
 * {SHA512}	SHA-512 without salt	[FIPS-180-4]<br/>
 * {SSHA512}	salted SHA-512	[FIPS-180-4]<br/>
 *
 */
public final class HashedPassword {

    public static final String SCHEME_CLEARTEXT = "CLEAR";
    public static final String SCHEME_MD5 = "MD5";
    public static final String SCHEME_SHA = "SHA";
    public static final String SCHEME_SHA256 = "SHA256";
    public static final String SCHEME_SHA384 = "SHA384";
    public static final String SCHEME_SHA512 = "SHA512";

    private static final Pattern SCHEME_PATTERN = Pattern.compile("\\{([A-Z0-9]+)\\}(\\S+)");
    private static final Pattern HEX_PATTERN = Pattern.compile("(?:[0-9A-Fa-f]{2})+");

    private final String scheme;
    private final byte[] hash;
    private final byte[] salt;

    private HashedPassword(String scheme, byte[] hash, byte[] salt) {
        this.scheme = scheme == null || scheme.isEmpty() ? SCHEME_CLEARTEXT : scheme;
        this.hash = hash == null ? new byte[0] : hash;
        this.salt = salt == null ? new byte[0] : salt;
    }

    public boolean isSalted() {
        return salt.length > 0;
    }

    public String getScheme() {
        return scheme;
    }

    public byte[] getSalt() {
        return salt;
    }

    /**
     * the hashed and (optional salted) password.
     */
    public byte[] getHash() {
        return hash;
    }

    @Override
    public int hashCode() {
        int offset = 5;
        offset = 97 * offset + Objects.hashCode(this.scheme);
        offset = 97 * offset + Arrays.hashCode(this.hash);
        offset = 97 * offset + Arrays.hashCode(this.salt);
        return offset;
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final HashedPassword other = (HashedPassword) obj;
        if (!Objects.equals(this.scheme, other.scheme)) {
            return false;
        }
        if (!Arrays.equals(this.hash, other.hash)) {
            return false;
        }
        return Arrays.equals(this.salt, other.salt);
    }

    @Override
    public String toString() {
        return encode();
    }

    /**
     * encodes the state as {<'S' if salted>&lt;SCHEME&gt;}&lt;BASE64 of
     * HASH&gt;
     */
    public String encode() {
        StringBuilder sb = new StringBuilder();
        encodeSchemeName(sb);
        byte[] hashAndSalt = concat(hash, salt);
        byte[] base64 = Base64.encodeBase64(hashAndSalt, false);
        sb.append(utf8BytesToString(base64));
        return sb.toString();
    }

    /**
     * Encodes the state as {&lt;SCHEME&gt;}&lt;HEX of HASH&gt;<br/>
     * Only for test purpose to check backwards compatibility.
     */
    String encodeAsHex() {
        return encodeAsHex(false);
    }

    /**
     * Encodes the state as &lt;HEX of HASH&gt;<br/>
     * Only for test purpose to check backwards compatibility.
     */
    String encodeAsHex(boolean skipScheme) {
        StringBuilder sb = new StringBuilder();
        if (!skipScheme) {
            encodeSchemeName(sb);
        }
        byte[] hashAndSalt = concat(hash, salt);
        sb.append(DigestUtil.digestToString(hashAndSalt));
        return sb.toString();
    }

    private void encodeSchemeName(StringBuilder sb) {
        sb.append("{");
        if (isSalted()) {
            sb.append("S");
        }
        sb.append(scheme);
        sb.append("}");
    }

    /**
     * Normalizes a scheme name into a formal spelling.
     */
    public static String normalizeScheme(String scheme) {
        return digestAlgorithmToScheme(schemeToDigestAlg(scheme));
    }

    /**
     * Tests if a scheme name is known and supported.
     */
    public static boolean isSupportedScheme(String scheme) {
        try {
            schemeToDigestAlg(scheme);
            return true;
        } catch (IllegalArgumentException ex) {
            return false;
        }
    }

    /**
     * Parses encoded hashed passwords.
     */
    public static HashedPassword fromString(String encodedString) {
        if (encodedString == null || encodedString.isEmpty()) {
            throw new IllegalArgumentException("encodedString is null or empty!");
        }
        if (!encodedString.startsWith("{")) {
            // clear text or old hash
            return parseFromNonQualifiedString(encodedString);
        }
        return parseFromQualifiedString(encodedString);
    }

    /**
     * creates hashed and salted passwords. The salt has a random length between
     * 16 and 48 bytes.
     */
    public static HashedPassword fromPassword(char[] password, String scheme) {
        // create a random salt
        return HashedPassword.fromPassword(password, scheme, SaltUtil.createSaltWithRandomLength());
    }

    /**
     * Creates a hashed and salted password based on the input parameters. Salt
     * can be null or empty array to create non salted passwords.
     */
    public static HashedPassword fromPassword(char[] password, String scheme, byte[] salt) {
        byte[] pw = charsToUTF8Bytes(password);
        MessageDigest digest = createDigestForScheme(scheme);
        if (digest == null) {
            // clear text can never be salted!
            return new HashedPassword(SCHEME_CLEARTEXT, pw, null);
        }
        byte[] saltedPw = concat(pw, salt);
        byte[] hashedSaltedPw = digest.digest(saltedPw);
        return new HashedPassword(digestAlgorithmToScheme(digest.getAlgorithm()), hashedSaltedPw, salt);
    }

    private static HashedPassword parseFromQualifiedString(String encodedString) {
        Matcher m = SCHEME_PATTERN.matcher(encodedString);
        if (!m.matches()) {
            throw new IllegalArgumentException("encoded strings does not match pattern '{<SCHEME>}base64'");
        }
        String scheme = m.group(1);
        String value = m.group(2);
        byte[] hashAndSalt = isHexString(value) ? hexToByteArray(value) : Base64.decodeBase64(value);
        MessageDigest digest = createDigestForScheme(scheme);
        if (digest == null) {
            // is clear text pw (never salted)
            return new HashedPassword(SCHEME_CLEARTEXT, hashAndSalt, null);
        }
        int l = digest.getDigestLength();
        byte[] hash = Arrays.copyOf(hashAndSalt, l);
        byte[] salt = Arrays.copyOfRange(hashAndSalt, l, hashAndSalt.length);
        return new HashedPassword(digestAlgorithmToScheme(digest.getAlgorithm()), hash, salt);
    }

    private static HashedPassword parseFromNonQualifiedString(String nonQualifiedSchemeString) {
        if (isHexString(nonQualifiedSchemeString)) {
            // old plain hash (salts not supported here)!
            byte[] hash = hexToByteArray(nonQualifiedSchemeString);
            String scheme = guessSchemeForHash(hash);
            if (scheme != null) {
                // guess success
                return new HashedPassword(scheme, hash, null);
            }
        }
        return new HashedPassword(SCHEME_CLEARTEXT, stringToUTF8Bytes(nonQualifiedSchemeString), null);
    }

    private static String guessSchemeForHash(byte[] hash) {
        // go though well known digests
        if (DigestUtil.isDigestOf(hash, schemeToDigestAlg(SCHEME_MD5))) {
            return SCHEME_MD5;
        }
        if (DigestUtil.isDigestOf(hash, schemeToDigestAlg(SCHEME_SHA))) {
            return SCHEME_SHA;
        }
        if (DigestUtil.isDigestOf(hash, schemeToDigestAlg(SCHEME_SHA256))) {
            return SCHEME_SHA256;
        }
        if (DigestUtil.isDigestOf(hash, schemeToDigestAlg(SCHEME_SHA384))) {
            return SCHEME_SHA384;
        }
        if (DigestUtil.isDigestOf(hash, schemeToDigestAlg(SCHEME_SHA512))) {
            return SCHEME_SHA512;
        }
        // guess failed
        return null;
    }

    private static String digestAlgorithmToScheme(String digestAlg) {
        switch (digestAlg) {
            case "MD5":
                return SCHEME_MD5;
            case "SHA":
                return SCHEME_SHA;
            case "SHA-1":
                return SCHEME_SHA;
            case "SHA-256":
                return SCHEME_SHA256;
            case "SHA-384":
                return SCHEME_SHA384;
            case "SHA-512":
                return SCHEME_SHA512;
            case SCHEME_CLEARTEXT:
                return SCHEME_CLEARTEXT;
            default:
                throw new IllegalArgumentException("message digest algorithm " + digestAlg + " not supported!");
        }
    }

    private static String schemeToDigestAlg(String scheme) {
        switch (scheme) {
            case "S" + SCHEME_MD5:
            case SCHEME_MD5:
                return "MD5";
            case "S" + SCHEME_SHA:
            case SCHEME_SHA:
            case "SHA-1":
                return "SHA-1";
            case "S" + SCHEME_SHA256:
            case SCHEME_SHA256:
            case "SHA-256":
                return "SHA-256";
            case "S" + SCHEME_SHA384:
            case SCHEME_SHA384:
            case "SHA-384":
                return "SHA-384";
            case "S" + SCHEME_SHA512:
            case SCHEME_SHA512:
            case "SHA-512":
                return "SHA-512";
            case SCHEME_CLEARTEXT:
                return SCHEME_CLEARTEXT;
            default:
                throw new IllegalArgumentException("scheme " + scheme + " is not supported!");
        }
    }

    private static MessageDigest createDigestForScheme(String scheme) {
        String alg = schemeToDigestAlg(scheme);
        if (SCHEME_CLEARTEXT.equals(alg)) {
            return null;
        }
        return DigestUtil.getMessageDigest(alg);
    }

    private static boolean isHexString(String value) {
        return HEX_PATTERN.matcher(value).matches();
    }

    private static byte[] hexToByteArray(String hashAsHex) {
        int len = hashAsHex.length();
        byte[] data = new byte[len / 2];
        for (int i = 0; i < len; i += 2) {
            data[i / 2] = (byte) ((Character.digit(hashAsHex.charAt(i), 16) << 4)
                    + Character.digit(hashAsHex.charAt(i + 1), 16));
        }
        return data;
    }

    private static byte[] concat(byte[] a, byte[] b) {
        if (a == null) {
            a = new byte[0];
        }
        if (b == null) {
            b = new byte[0];
        }
        byte[] out = new byte[a.length + b.length];
        System.arraycopy(a, 0, out, 0, a.length);
        System.arraycopy(b, 0, out, a.length, b.length);
        return out;
    }

    private static String utf8BytesToString(byte[] base64) {
        try {
            return new String(base64, "UTF-8");
        } catch (UnsupportedEncodingException ex) {
            throw new IllegalStateException("UTF-8 not supported!", ex);
        }
    }

    private static byte[] charsToUTF8Bytes(char[] password) throws IllegalStateException {
        return stringToUTF8Bytes(new String(password));
    }

    private static byte[] stringToUTF8Bytes(String value) throws IllegalStateException {
        try {
            return value.getBytes("UTF-8");
        } catch (UnsupportedEncodingException ex) {
            throw new IllegalStateException("UTF-8 not supported", ex);
        }
    }
}
