/*******************************************************************************
 * Copyright (C) 2009
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 ******************************************************************************/

package org.n52.security.common.subject;

import java.io.Serializable;
import java.security.Principal;

/**
 * Minimal implementation of a principal.
 * It is declared abstract to avoid the usage of such unspecific principals.
 * <p/>
 * A login module shall use one of its subclasses for better application support.
 * <p/>
 * To support better differentiation of principals a scope is introduced, this can be used to distinguish principals
 * with the same name and class.
 *
 * @author Marko Reiprecht
 *         Created on 16.09.2007
 * @version $Revision: $
 */
public abstract class NamePrincipal implements Principal, Scope, Serializable {

    private static final long serialVersionUID = 4569056343464389737L;

    private String m_name;

    private Scope m_scope;

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * Constructor NamePrincipal creates a new NamePrincipal instance,
     * which name is only valid within the given scope.
     * <p/>
     * It throws an IllegalArgumentException if one of the arguments is null.
     *
     * @param name
     *            the name != null.
     * @param scope
     *            the scope != null.
     * @throws IllegalArgumentException
     *             if one of the arguments is null.
     */
    public NamePrincipal(final String name, final Scope scope) throws IllegalArgumentException {
        if (name == null) {
            throw new IllegalArgumentException("the name of a named principal must not null");
        }
        if (scope == null) {
            throw new IllegalArgumentException("the scope of a named principal must not null");
        }
        m_name = name;
        m_scope = scope;
    }

    public String getName() {
        return m_name;
    }

    public Scope getScope() {
        return m_scope;
    }

    public boolean equals(final Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        final NamePrincipal that = (NamePrincipal) o;

        if (!m_name.equals(that.m_name)) {
            return false;
        }
        if (!m_scope.equals(that.m_scope)) {
            return false;
        }

        return true;
    }

    public int hashCode() {
        int result;
        result = m_name.hashCode();
        result = 31 * result + m_scope.hashCode();
        return result;
    }

    public String toString() {
        StringBuffer builder = new StringBuffer();
        builder.append("NamePrincipal [name=");
        builder.append(m_name);
        builder.append(", scope=");
        builder.append(m_scope);
        builder.append("]");
        return builder.toString();
    }
}
