package org.n52.security.common.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Very simple implementation of a ObjectPool.
 */
public abstract class AbstractObjectPool<PooledObject, Key> implements ObjectPool<PooledObject, Key> {
    private static final org.slf4j.Logger LOG = org.slf4j.LoggerFactory.getLogger(AbstractObjectPool.class);
    private boolean m_destroyed;
    private Map<Key, LinkedList<PooledObject>> m_free = new HashMap<Key, LinkedList<PooledObject>>();
    private Map<Key, List<PooledObject>> m_inUse = new HashMap<Key, List<PooledObject>>();
    private int m_cachedLimit;

    public AbstractObjectPool(int cachedObjLimit) {
        m_cachedLimit = cachedObjLimit;
    }

    @Override
    public synchronized PooledObject acquire(Key key) {
        if (m_destroyed) {
            throw new IllegalStateException(
                    "<acquire> is not allowed to be called after destruction of the pool! Key: " + key);
        }
        PooledObject pooledObj = getFromFree(key);
        return addToUse(pooledObj == null ? create(key) : pooledObj, key);
    }

    private PooledObject getFromFree(Key key) {
        LinkedList<PooledObject> freeObjs = m_free.get(key);
        if (freeObjs != null && freeObjs.size() > 0) {
            return freeObjs.removeFirst();
        }
        return null;
    }

    private PooledObject addToUse(PooledObject obj, Key key) {
        List<PooledObject> usedObjs = m_inUse.get(key);
        if (usedObjs == null) {
            usedObjs = new ArrayList<PooledObject>();
            m_inUse.put(key, usedObjs);
        }
        usedObjs.add(obj);
        LOG.trace("in use are <{}> objects of type <{}>", usedObjs.size(), obj.getClass());
        return obj;
    }

    @Override
    public synchronized void release(PooledObject obj, Key key) {
        if (m_destroyed) {
            return;
        }
        removeFromUse(obj, key);
        addToFree(obj, key);
    }

    private void addToFree(PooledObject obj, Key key) {
        LinkedList<PooledObject> freeObjs = m_free.get(key);
        if (freeObjs == null) {
            freeObjs = new LinkedList<PooledObject>();
            m_free.put(key, freeObjs);
        }
        if (freeObjs.size() > m_cachedLimit) {
            destroy(key, obj);
            return;
        }
        freeObjs.addFirst(obj);
        LOG.trace("free are <{}> objects of type <{}>", freeObjs.size(), obj.getClass());
    }

    private void removeFromUse(PooledObject obj, Key key) {
        List<PooledObject> objsInUse = m_inUse.get(key);
        if (objsInUse != null) {
            objsInUse.remove(obj);
        }
    }

    public void destroy(Key key, PooledObject obj) {
        // nothing by default
    }

    public synchronized void destroy() {
        // destroys the full pool , also the objects in use
        if (!m_destroyed) {
            m_destroyed = true;
            destroy(m_free);
            destroy(m_inUse);
        }
    }

    private void destroy(Map<Key, ? extends List<PooledObject>> pool) {
        for (Key key : pool.keySet()) {
            for (PooledObject obj : pool.get(key)) {
                destroy(key, obj);
            }
        }
    }

    /**
     * only create needs to be implemented by sub classes.
     */
    public abstract PooledObject create(Key key);

}
