/** *****************************************************************************
 * Copyright (C) 2008 by 52 North Initiative for Geospatial Open Source Software
 * GmbH
 *
 * Contact: Andreas Wytzisk 52 North Initiative for Geospatial Open Source
 * Software GmbH Martin-Luther-King-Weg 24 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 ***************************************************************************** */
package org.n52.security.common.util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;

/**
 * path parts of an url.
 */
public final class URLPath {

    private static final String PATH_DELIMITER = "/";
    private static final URLPath ROOT = new URLPath("/");
    private static final URLPath EMPTY = new URLPath("");

    public static URLPath fromString(String... parts) {
        URLPath path = new URLPath(parts);
        if (path.isRoot()) {
            return ROOT;
        }
        return path;
    }

    public static URLPath root() {
        return ROOT;
    }

    public static URLPath empty() {
        return EMPTY;
    }

    private final List<String> segments;

    /**
     * creates a new {@link UrlPath} from an array of elements
     *
     * @param segments the array of elements
     * @throws IllegalArgumentException if one or more elements couldn't parsed
     */
    private URLPath(String... segments) throws IllegalArgumentException {
        this.segments = parse(segments);
    }

    private URLPath(List segments) {
        this.segments = new ArrayList<>(segments);
    }

    /**
     * gives an array of all elements
     *
     * @return an array of all elements
     */
    public List<String> getSegments() {
        return Collections.unmodifiableList(segments);
    }

    public String getSegment(int i) {
        return segments.get(i);
    }

    public String getSegmentOrEmptyString(int i) {
        return i > 0 && i < size() ? segments.get(i) : "";
    }

    /**
     * constructs the string representation of the path.
     *
     * @return a string.
     */
    public String asString() {
        if (size() == 1 && hasTrailingSlash()) {
            // special case if only one part, then join will not append the delimter.
            return PATH_DELIMITER;
        }
        return StringUtils.join(segments, PATH_DELIMITER);
    }

    /**
     * gives the count of elements
     *
     * @return the number of elements
     */
    public int size() {
        return segments.size();
    }

    /**
     * checks, if this {@link URLPath} have no elements
     *
     * @return {@code true} if this {@link URLPath} have no elements, otherwise
     * {@code false}
     */
    public boolean isEmpty() {
        return EMPTY.equals(this);
    }

    /**
     * if this path represents a path root (single slash);
     */
    public boolean isRoot() {
        return ROOT.equals(this);
    }

    /**
     * gives an {@link URLPath}, which contains all elements starting from the
     * given index.
     *
     * @param start of the first element (including) {@link URLPath}
     * @return the {@link URLPath} with the sub elements
     */
    public URLPath getSubPath(int start) {
        return getSubPath(start, size());
    }

    /**
     * gives an {@link URLPath}, which contains all elements in the given range.
     *
     * @param start index (including)
     * @param end index (excluding)
     * @return the {@link URLPath} with the sub elements
     */
    public URLPath getSubPath(int start, int end) {
        List<String> parts = getSegments();
        List<String> newSegments = parts.subList(start, end);
        URLPath newPath = new URLPath(newSegments);
        if (newPath.isEmpty()) {
            return EMPTY;
        }
        if (newPath.isRoot()) {
            return ROOT;
        }
        return newPath;
    }

    /**
     * Appends the given url path to this path.
     */
    public URLPath append(URLPath toAppend) {
        if (toAppend == null || toAppend.isEmpty()) {
            return this;
        }
        if (this.isEmpty()) {
            return toAppend;
        }
        if (this.isRoot() && toAppend.isRoot()) {
            return ROOT;
        }
        ArrayList newParts = new ArrayList(removeTrailingSlash().segments);
        newParts.addAll(toAppend.segments);
        return new URLPath(newParts);
    }

    /**
     * checks, if this {@link URLPath} ends with a slash
     *
     * @return {@code true} if it contains a trailing slash, otherwise
     * {@code false}
     */
    public boolean hasTrailingSlash() {
        if (isEmpty()) {
            return false;
        }
        return getSegment(size() - 1).isEmpty();
    }

    /**
     * gives a {@link URLPath} without a trailing slash
     *
     * @return a new created {@link URLPath}, which have all elements, without a
     * trailing slash
     */
    public URLPath removeTrailingSlash() {
        if (!hasTrailingSlash()) {
            return this;
        }
        return getSubPath(0, size() - 1);
    }

    /**
     * gives a {@link URLPath} with a trailing slash
     *
     * @return a new created {@link URLPath}, which have all elements, with a
     * trailing slash
     */
    public URLPath appendTrailingSlash() {
        if (hasTrailingSlash()) {
            return this;
        }
        return append(ROOT);
    }

    @Override
    public String toString() {
        return asString();
    }

    @Override
    public int hashCode() {
        int hash = 7;
        hash = 29 * hash + Objects.hashCode(this.segments);
        return hash;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final URLPath other = (URLPath) obj;
        if (!Objects.equals(this.segments, other.segments)) {
            return false;
        }
        return true;
    }

    private List<String> parse(String[] segments) {
        List<String> out = new ArrayList<>();
        // parse recursive and skip empty parts
        parseNonEmptySegments(out, segments);
        // check for special trailing "/" condition and parameter construction
        String lastSegment = segments != null && segments.length > 0 ? segments[segments.length - 1] : null;
        if (lastSegment != null
                && (lastSegment.endsWith("/")
                || (lastSegment.isEmpty() && out.size() > 0))) {
            // push empty segment to end
            out.add("");
        }
        return out;
    }

    private void parseNonEmptySegments(List<String> out, String... elements) {
        // intercept empty lists
        if (elements == null || elements.length == 0) {
            return;
        }

        for (String element : elements) {
            String[] parts = element.split(PATH_DELIMITER);
            if (parts.length > 1) {
                parseNonEmptySegments(out, parts);
            } else if (parts.length == 1) {
                parseNonEmptySegment(out, parts[0]);
            }
        }
    }

    private void parseNonEmptySegment(List<String> out, String segment) {
        if (segment == null || segment.isEmpty()) {
            return;
        }
        out.add(segment);
    }
}
