/*
 *  Copyright (C) con terra GmbH (http://www.conterra.de)
 *  All rights reserved
 */
package org.n52.security.common.xml;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.ref.WeakReference;
import java.net.URL;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xml.sax.InputSource;

public class ClasspathResourceResolver {

    private static final Logger LOG = LoggerFactory.getLogger(ClasspathResourceResolver.class);
    private static final byte[] EMPTY_BYTES = new byte[0];
    private static final String MAPPING_FILE_LOCATION = "META-INF/xml/entity-mapping.properties";
    private static Map<String, String> DEFAULT_MAPPING = null;

    private Map<String, String> m_systemToClasspathMapping = new HashMap<>();
    private WeakReference<ClassLoader> m_classLoader;
    private boolean m_quiet;

    public ClasspathResourceResolver() {
        this(true, null);
    }

    public ClasspathResourceResolver(final boolean quiet) {
        this(quiet, null);
    }

    public ClasspathResourceResolver(final boolean quiet, ClassLoader loader) {
        m_quiet = quiet;
        loader = loader == null ? this.getClass().getClassLoader() : loader;
        m_systemToClasspathMapping = loadMapping(loader);
        m_classLoader = new WeakReference<>(loader);
    }

    public void setMapping(Map<String, String> mapping) {
        this.m_systemToClasspathMapping = mapping == null ? new HashMap<String, String>() : mapping;
    }

    public InputSource lookup(String id) {
        InputSource result = lookupIgnoreQuite(id);
        if (result == null && isQuite()) {
            return emptyInputSource();
        }
        return result;
    }

    public InputSource lookupIgnoreQuite(String id) {
        if (id == null || id.isEmpty()) {
            return null;
        }
        String resourcefile = getMappedResourceFileName(id);
        if (resourcefile == null) {
            return null;
        }
        return createInputSource(resourcefile);
    }

    /**
     * Method loads all entries from the mapping files located at
     * META-INF/xml/entity-mapping.properties. A typical entity mapping looks
     * like this:
     * <p/>
     * <code>[publicId|systemId]=path in classpath</code>
     * <p/>
     * e.g.:
     * <p/>
     * <code>http://tempuri.org/test.dtd=META-INF/xml/dtds/test.dtd</code>
     *
     * @param classloader the classloader if null
     * this.getClass().getClassLoader() is used.
     */
    private static Map<String, String> loadMapping(final ClassLoader resourceLoader) {
        if (DEFAULT_MAPPING != null && resourceLoader == ClasspathResourceResolver.class.getClassLoader()) {
            return new HashMap<>(DEFAULT_MAPPING);
        }
        HashMap<String, String> resolvedMappings = new HashMap<>();
        try {
            Enumeration files = resourceLoader.getResources(MAPPING_FILE_LOCATION);
            while (files.hasMoreElements()) {
                URL resource = (URL) files.nextElement();
                Properties mappings = new Properties();
                InputStream in = resource.openStream();
                try {
                    mappings.load(resource.openStream());
                    for (String key : mappings.stringPropertyNames()) {
                        resolvedMappings.put(key, mappings.getProperty(key));
                    }
                } catch (IOException ex) {
                    if (LOG.isWarnEnabled()) {
                        LOG.warn("could not load <" + resource + ">. " + "Loading of entities specified within this mapping file will fail. Error reason:", ex);
                    }
                } finally {
                    try {
                        in.close();
                    } catch (IOException ex) {
                        // TODO: ignore?
                    }
                }
            }
        } catch (IOException e) {
            if (LOG.isWarnEnabled()) {
                LOG.warn("could not load <entity-mapping.properties> from classpath, " + "loading of entities will fail. Error reason:", e);
            }
        }
        if (DEFAULT_MAPPING == null && resourceLoader == ClasspathResourceResolver.class.getClassLoader()) {
            return new HashMap<>(DEFAULT_MAPPING = resolvedMappings);
        }
        return resolvedMappings;
    }

    /**
     * Method gets the quiet state. If the EnityReslover is in quiet mode, all
     * non resolved entities are simply ignored (not loaded), otherwise the
     * parsers default mode is used.
     *
     * @return boolean.
     */
    public boolean isQuite() {
        return m_quiet;
    }

    /**
     * Method creates an input source for the given resource file.
     * <p/>
     * It tries to load the resourcefile from the classpath. If it fails it
     * simple ignores the entity.
     *
     * @param resourcefile the resourcefile.
     * @return InputSource.
     */
    private InputSource createInputSource(final String resourcefile) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Try to resolve entity from resource file <" + resourcefile + ">");
        }
        ClassLoader loader = (ClassLoader) m_classLoader.get();
        // if classloader was garbaged, the resource can't be loaded
        if (loader != null && resourcefile != null) {
            InputStream resourcestream = loader.getResourceAsStream(resourcefile);
            if (resourcestream != null) {
                return new InputSource(resourcestream);
            } else if (LOG.isWarnEnabled()) {
                LOG.warn("entity mapping with value <" + resourcefile + "> seems broken," + " because the resourcestream does not exist in the classpath." + " Entity won't be resolved.");
            }
        }
        return null;
    }

    /**
     * Method creates an InputSource which points to an empty resource.
     *
     * @return InputSource.
     */
    protected InputSource emptyInputSource() {
        return new InputSource(new ByteArrayInputStream(EMPTY_BYTES));
    }

    /**
     * Method gets the mapped resource file name for the given identifier.
     *
     * @param identifier the identifier (e.g. publicId or systemId).
     * @return String if mapping is found, otherwise null.
     */
    private String getMappedResourceFileName(final String identifier) {
        return m_systemToClasspathMapping.get(identifier);
    }

}
