/*******************************************************************************
 * Copyright (C) 2008
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 * Author: Marko Reiprecht
 * Created: 18.04.2008
 ******************************************************************************/
package org.n52.security.common.xml;

import java.util.List;

import javax.xml.namespace.QName;

/**
 * Processes XML chunks.
 * <p/>
 * XML chunks are valid XML documents which were part of an greater XML document.
 * <p/>
 * The interface is directly related to a XMLChunkReader, which reads an big xml document and calls an XMLChunkHandler
 * to process some of it's parts. A XMLChunkHandler can parse the chunk with an DOMParser or an other XMLprocessing
 * tool.
 * <p/>
 * If an XMLChunkHandler is responsible for a special xml element and within it's scope other XMLChunkHandlers are
 * registered, then the handler is informed about the processing of the childs through the *Child* methods.
 *
 * @author Marko Reiprecht
 *         Created on 18.04.2008
 * @version $Revision: $
 */
public interface XMLChunkHandler {

    void handleChunkStart(final List<QName> currentXPath, final QName handlerQName);

    /**
     * Invoked if the end tag of the xml block/chunk is reached.
     *
     * @param currentXPath
     *            Xpath (QName Path) to the current element.
     * @param handlerQName
     *            the handlerQName.
     * @param xmlChunk
     *            the partialXml.
     */
    void handleChunkEnd(final List<QName> currentXPath, final QName handlerQName, final String xmlChunk);

    /**
     * Invoked if the start tag of an element, for which a child handler is registered
     * is reached.
     * The xmlChunk is not well formed, because the end tag is not closed.
     *
     * @param currentXPath
     *            Xpath (QName Path) to the current element.
     * @param handlerQName
     *            the handlerQName.
     * @param xmlChunk
     *            the xmlChunk.
     * @param childQName
     *            the childQName.
     * @param childHandler
     *            the childHandler.
     */
    void handleChunkBeforChildHandler(final List<QName> currentXPath,
                                      final QName handlerQName,
                                      final String xmlChunk,
                                      final QName childQName,
                                      final XMLChunkHandler childHandler);

    /**
     * Invoked if the end tag of an element, for which a child handler is registered is reached.
     * <p/>
     * The xmlChunk is not well formed, because the end tag is not closed.
     *
     * @param currentXPath
     *            Xpath (QName Path) to the current element.
     * @param handlerQName
     *            the handlerQName.
     * @param xmlChunk
     *            the xmlChunk.
     * @param childQName
     *            the childQName.
     * @param childHandler
     *            the childHandler.
     */
    void handleChunkAfterChildHandler(final List<QName> currentXPath,
                                      final QName handlerQName, final String xmlChunk,
                                      final QName childQName, final XMLChunkHandler childHandler);

}
